import math
import warnings
import numpy as np
import geopandas as gpd
from scipy.spatial import cKDTree


def _get_coords(gdf, method="random", size=1, batch_size=1000):
    """
    Get array containing x,y coordinates from GeoDataFrame

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        GeoDataFrame to return coordinates from
    method : str
        what method to use to sample polygons
    size : int (default: 1)
        number of random points to generate within polygon (default 1)
    batch_size : int (default: 1000)
        number of random points to generate within bounds

    Returns
    -------
    type : np.array
        array containing x and y values
    """

    geom_type = gdf.type.values

    if "x" and "y" in gdf:
        return gdf[["x", "y"]].values

    else:
        if geom_type.any() == "Point":
            x, y = gdf.geometry.x, gdf.geometry.y
            return np.array([x, y]).T

        elif geom_type.any() in ("Polygon", "MultiPolygon"):
            if method == "random":
                return sample_points(gdf, size, batch_size)
            else:
                x_mids, y_mids = _calculate_mid_points(gdf.bounds)
                return np.array([x_mids, y_mids]).T

        elif geom_type.any() in ("LineString", "MultiLineString"):
            raise ValueError("Not suppported")


def sample_points(gdf, size, batch_size):
    """
    Get array containing x,y coordinates from GeoDataFrame

    Parameters
    ----------
    gdf : gpd.GeoDataFrame

    size : int (default: 1)
        number of random points to generate within polygon (default 1)
    batch_size : int (default: 1000)
        number of random points to generate within bounds

    Returns
    -------
    type : np.array
        array containing x and y values
    """

    geoms = gdf.geometry.apply(_uniform_polygon, size=size)

    g = gpd.GeoDataFrame(geometry=geoms, crs=gdf.crs)

    g["x"] = g.geometry.x
    g["y"] = g.geometry.y

    return g[["x", "y"]].values


def _uniform_polygon(geom, size=1, batch_size=None):
    """
    Return n points contained within a polygon

    Parameters
    ----------
    geom : gpd.GeoDataFrame.geometry
        geometry for a gpd.GeoDataFrame
    size : int (default: 1)
        number of random points to generate within polygon (default 1)
    batch_size : int (default: 1000)
        number of random points to generate within bounds

    Returns
    -------
    type : gpd.GeoSeries
        GeoSeries containing n points contained within a polygon
    """
    from geopandas.array import points_from_xy
    from geopandas.geoseries import GeoSeries

    n_points = size

    if batch_size is None:
        batch_size = n_points

    xmin, ymin, xmax, ymax = geom.bounds

    candidates = []

    while len(candidates) < n_points:
        batch = points_from_xy(
            x=np.random.uniform(xmin, xmax, size=batch_size),
            y=np.random.uniform(ymin, ymax, size=batch_size),
        )
        valid_samples = batch[batch.sindex.query(geom, predicate="contains")]
        candidates.extend(valid_samples)

    return GeoSeries(candidates[:n_points]).unary_union


def _calculate_mid_points(bounds):
    """
    Calculate middle points based on the geometry bounds

    Parameters
    ----------
    bounds : array_like
        array containing xmin, ymin, xmax, ymax

    Returns
    -------
    type: np.array
        x_mids : mid points of x values
        y_mids : mid points of y values
    """

    # Calculate mid points for x and y bound coords
    x_mids = (bounds[:, 0] + bounds[:, 2]) / 2.0
    y_mids = (bounds[:, 1] + bounds[:, 3]) / 2.0

    return x_mids, y_mids


def _check_bandwidth(coords1, coords2, bandwidth):

    """A function that checks whether the bandwidth set
    allows for pooling from at least 1 neighbour

    Parameters
    ----------
    coords1: np.array
        array containing x,y coordinate pairs
    coords2: np.array
        array containing x,y coordinate pairs
    bandwidth: int
        fixed distance for finding neighbours

    Returns
    -------
    int: (optional)
        bandwidth

    """

    # Create KD-tree
    tree = cKDTree(coords1)
    # Get number of neighbours within bandwidth
    k = tree.query_ball_point(coords2, r=bandwidth, return_length=True)
    # If any observation has less than 1 neighbour,
    # then calculate the minimum distance for each to have neighbours
    if np.any(k <= 1) is True:
        d, i = tree.query(coords2, k=1)
        max_d = math.ceil(d.max())

        warnings.warn(
            f"The bandwidth: {bandwidth} will result in the indexes:"
            f"{i} not having any values pooled. If this is not desired, use"
            f"a bandwidth greater than {max_d}"
        )



from scipy.spatial.distance import cdist
from geobootstrap.utils import _get_coords, _check_bandwidth
from geobootstrap.kernel import _kernel


def geobootstrap(
    gdf1,
    gdf2,
    coords1=None,
    coords2=None,
    r=1000,
    kernel="gaussian",
    metric="euclidean",
    bandwidth=1000,
    fixed=True,
):
    """
    Bootstrap a GeoDataFrame using defined kernel weights,
    determined by a distance decay function to another GeoDataFrame

    Parameters
    ----------
    gdf1 : gpd.GeoDataFrame
        GeoDataFrame to calculate distances from
    gdf2 : gpd.GeoDataFrame
        GeoDataFrame to calculate distances to
    coords1: array_like
        two dimensional array containing x, y values for sources
   coords2: array_like
        two dimensional array containing x, y values for gdf2
    r : int
        how many resamples with replacement to return
    kernel : str
        kernel distance-decay function
    metric : str
        how to calculate distances between coordinates
    bandwidth : int
       bandwidth (distance)
    fixed: bool
        whether to apply a fixed or adaptive (knn) kernel

    Returns
    -------
    type : list
        list of GeoPandas.GeoDataFrames
    """

    if fixed is False:
        raise ValueError("Only fixed bandwidths are currently supported")

    if coords1 is None:
        coords1 = _get_coords(gdf1)
    if coords2 is None:
        coords2 = _get_coords(gdf2)

    # Check bw if all observations will have at least 1 neighbour
    _check_bandwidth(coords1, coords2, bandwidth)
    # Compute distances between coords
    dist = cdist(coords2, coords1, metric)
    # Calculate kernel for each gdf2 observation
    ks = _kernel(kernel, dist, bandwidth, fixed)

    return [gdf1.sample(n=r, weights=k, replace=True) for k in ks]



import numpy as np
import warnings


def _kernel(function, dist, bandwidth, fixed=True):
    """
    Creates distanced-based kernel weights for geobootstrap

    In general, a fixed bandwidth suits more regular sample configurations,
    whereas an adaptive bandwidth suits irregular sample configurations

    Parameters
    ----------
    function : str
        Kernel function specified
    dist : array_like
        Distance vector
    bandwidth : int
        fixed bandwidth in metres
    fixed : bool
        whether to return a fixed or adaptive bandwidth

    Returns
    -------
    type: array_like
        distance-based kernel weights
    """

    if fixed is True:
        return _kernel_funcs(function, zs=dist / bandwidth)


def _kernel_funcs(function, zs):
    """
    modified from
    https://github.com/pysal/mgwr/blob/eef4f707fc2d34245c6cf0afeb85d742fdb90af5/mgwr/kernels.py#L65

    Parameters
    ----------
    function : str
        kernel function name
    zs : int
        distance / bandwidth

    Returns
    -------
    type: array_like
        kernel weights
    """

    functions = [
        "triangular",
        "uniform",
        "quadratic",
        "quartic",
        "gaussian",
        "bisquare",
        "exponential",
    ]

    if function != "gaussian":
        warnings.warn("Only Gaussian kernels have been tested")

    if function == "triangular":
        return 1 - zs
    elif function == "uniform":
        return np.ones(zs.shape) * 0.5
    elif function == "quadratic":
        return (3.0 / 4) * (1 - zs ** 2)
    elif function == "quartic":
        return (15.0 / 16) * (1 - zs ** 2) ** 2
    elif function == "gaussian":
        return np.exp(-0.5 * (zs) ** 2)
    elif function == "bisquare":
        return (1 - (zs) ** 2) ** 2
    elif function == "exponential":
        return np.exp(-zs)
    else:
        raise ValueError(
            f"Only the following kernels are supported: \
            {functions}"
        )



from setuptools import setup, find_packages


def readme():
    with open("README.md") as f:
        return f.read()


install_requires = ["geopandas==0.10.2", "scipy==1.7.3"]

setup(
    name="geobootstrap",
    version="0.11",
    description="Python tools for geobootstrapping",
    long_description=(readme()),
    long_description_content_type="text/markdown",
    url="https://github.com/tastatham/geobootstrap",
    author="Thomas Statham",
    author_email="tastatham@gmail.com",
    keywords="bootstrap, spatial",
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
    include_package_data=False,
)
